import copy
import enum
import hashlib
import os
import pathlib
import shutil
import subprocess

import sphinx
from docutils.nodes import (
    Element,
    FixedTextElement,
    General,
    SkipNode,
    image,
    literal_block,
)
from docutils.parsers.rst import Directive, directives
from sphinx.util import logging
from sphinx.util.console import brown  # type: ignore
from sphinx.util.nodes import set_source_info
from sphinx.util.osutil import copyfile, ensuredir
from uqbar.book.extensions import Extension
from uqbar.strings import normalize

from .. import format as _format
from .. import lilypondfile as _lilypondfile
from .. import tag as _tag
from ..configuration import Configuration
from ..contextmanagers import TemporaryDirectoryChange
from ..illustrators import illustrate
from ..io import Illustrator, LilyPondIO, Player

configuration = Configuration()
logger = logging.getLogger(__name__)


class HiddenDoctestDirective(Directive):
    """
    A hidden doctest directive.

    Contributes no formatting to documents built by Sphinx.
    """

    ### CLASS VARIABLES ###

    __documentation_ignore_inherited__ = True

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec: dict[str, str] = {}

    ### PUBLIC METHODS ###

    def run(self):
        """Executes the directive."""
        self.assert_has_content()
        return []


class ShellDirective(Directive):
    """
    A shell directive.

    Represents a shell session.

    Generates a docutils ``literal_block`` node.
    """

    ### CLASS VARIABLES ###

    __documentation_ignore_inherited__ = True

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec: dict[str, str] = {}

    ### PUBLIC METHODS ###

    def run(self):
        self.assert_has_content()
        result = []
        with TemporaryDirectoryChange(configuration.abjad_directory):
            cwd = pathlib.Path.cwd()
            for line in self.content:
                result.append(f"{cwd.name}$ {line}")
                completed_process = subprocess.run(
                    line,
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    text=True,
                )
                result.append(completed_process.stdout)
        code = "\n".join(result)
        literal = literal_block(code, code)
        literal["language"] = "console"
        set_source_info(self, literal)
        return [literal]


class ThumbnailDirective(Directive):
    """
    A thumbnail directive.
    """

    ### CLASS VARIABLES ###

    __documentation_ignore_inherited__ = True

    final_argument_whitespace = True
    has_content = False
    option_spec = {
        "class": directives.class_option,
        "group": directives.unchanged,
        "title": directives.unchanged,
    }
    optional_arguments = 0
    required_arguments = 1

    ### PUBLIC METHODS ###

    def run(self):
        """Executes the directive."""
        node = thumbnail_block()
        node["classes"] += self.options.get("class", "")
        node["group"] = self.options.get("group", "")
        node["title"] = self.options.get("title", "")
        node["uri"] = self.arguments[0]
        environment = self.state.document.settings.env
        name, suffix = os.path.splitext(node["uri"])
        environment.images.add_file(environment.docname, node["uri"])
        thumbnail_uri = name + "-thumbnail" + suffix
        environment.thumbnails[thumbnail_uri] = environment.docname
        # this may also work:
        # environment.thumbnails[environment.docname] = thumbnail_uri
        return [node]


class thumbnail_block(image, General, Element):
    __documentation_ignore_inherited__ = True


def visit_thumbnail_block_html(self, node):
    template = normalize(
        """
        <a data-lightbox="{group}" href="{target_path}" title="{title}" data-title="{title}" class="{cls}">
            <img src="{thumbnail_path}" alt="{alt}"/>
        </a>
        """
    )
    title = node["title"]
    classes = " ".join(node["classes"])
    group = "group-{}".format(node["group"] if node["group"] else node["uri"])
    if node["uri"] in self.builder.images:
        node["uri"] = os.path.join(
            self.builder.imgpath, self.builder.images[node["uri"]]
        )
    target_path = node["uri"]
    prefix, suffix = os.path.splitext(target_path)
    if suffix == ".svg":
        thumbnail_path = target_path
    else:
        thumbnail_path = "{}-thumbnail{}".format(prefix, suffix)
    output = template.format(
        alt=title,
        group=group,
        target_path=target_path,
        cls=classes,
        thumbnail_path=thumbnail_path,
        title=title,
    )
    self.body.append(output)
    raise SkipNode


def visit_thumbnail_block_latex(self, node):
    raise SkipNode


def on_builder_inited(app):
    app.env.thumbnails = {}  # separate so Sphinx doesn't purge it
    install_lightbox_static_files(app)
    (pathlib.Path(app.builder.outdir) / "_images").mkdir(exist_ok=True)


class LilyPondExtension(Extension):
    class Kind(enum.Enum):
        IMAGE = 1
        AUDIO = 2

    class lilypond_block(General, FixedTextElement):
        pass

    @classmethod
    def setup_console(cls, console, monkeypatch):
        monkeypatch.setattr(
            Illustrator,
            "__call__",
            lambda self: console.push_proxy(
                cls(
                    self.illustrable,
                    cls.Kind.IMAGE,
                    **{
                        key.replace("lilypond/", "").replace("-", "_"): value
                        for key, value in console.proxy_options.items()
                        if key.startswith("lilypond/")
                    },
                ),
            ),
        )
        monkeypatch.setattr(
            Player,
            "__call__",
            lambda self: console.push_proxy(
                cls(
                    self.illustrable,
                    cls.Kind.AUDIO,
                    **{
                        key.replace("lilypond/", "").replace("-", "_"): value
                        for key, value in console.proxy_options.items()
                        if key.startswith("lilypond/")
                    },
                ),
            ),
        )

    @classmethod
    def setup_sphinx(cls, app):
        app.add_node(
            cls.lilypond_block,
            html=[cls.visit_block_html, None],
            latex=[cls.visit_block_latex, None],
            text=[cls.visit_block_text, cls.depart_block_text],
        )
        cls.add_option("lilypond/no-trim", directives.flag)
        cls.add_option("lilypond/pages", directives.unchanged)
        cls.add_option("lilypond/with-columns", int)

    def __init__(
        self,
        illustrable,
        kind,
        no_trim=None,
        pages=None,
        with_columns=None,
        **keywords,
    ):
        self.illustrable = copy.deepcopy(illustrable)
        self.keywords = keywords
        self.kind = kind
        self.no_trim = no_trim
        self.pages = pages
        self.with_columns = with_columns

    def to_docutils(self):
        if isinstance(self.illustrable, _lilypondfile.LilyPondFile):
            illustration = self.illustrable
        else:
            illustration = illustrate(self.illustrable, **self.keywords)
        if self.kind == self.Kind.AUDIO:
            block = _lilypondfile.Block("midi")
            illustration["score"].items.append(block)
        illustration.lilypond_version_token = r'\version "2.19.83"'
        code = illustration._get_lilypond_format()
        code = _format.remove_site_comments(code)
        code = _tag.remove_tags(code)
        node = self.lilypond_block(code, code)
        node["kind"] = self.kind.name.lower()
        node["no-trim"] = self.no_trim
        node["pages"] = self.pages
        node["with-columns"] = self.with_columns
        return [node]

    @staticmethod
    def visit_block_html(self, node):
        output_directory = pathlib.Path(self.builder.outdir) / "_images"
        render_prefix = "lilypond-{}".format(
            hashlib.sha256(node[0].encode()).hexdigest()
        )
        if node["kind"] == "audio":
            flags = []
            glob = f"{render_prefix}.mid*"
        else:
            flags = ["-dcrop", "-dbackend=svg"]
            glob = f"{render_prefix}*.svg"
        lilypond_io = LilyPondIO(
            None,
            flags=flags,
            output_directory=output_directory,
            render_prefix=render_prefix,
            should_copy_stylesheets=True,
            should_open=False,
            should_persist_log=False,
            string=node[0],
        )
        if not list(output_directory.glob(glob)):
            _, _, _, success, log = lilypond_io()
            if not success:
                logger.warning(f"LilyPond render failed\n{log}", location=node)
        source_path = (pathlib.Path(self.builder.imgpath) / render_prefix).with_suffix(
            ".ly"
        )
        if node["kind"] == "audio":
            pass
        else:
            embed_images(self, node, output_directory, render_prefix, source_path)
        raise SkipNode


table_row_open_template = '<div class="table-row">'
table_row_close_template = "</div>"
basic_image_template = normalize(
    """
    <div class="uqbar-book">
        <a href="{source_path}"><img src="{relative_path}"/></a>
    </div>
    """
)
thumbnail_template = normalize(
    """
    <a data-lightbox="{group}" href="{fullsize_path}" title="{title}" data-title="{title}" class="{cls}">
        <img src="{thumbnail_path}" alt="{alt}"/>
    </a>
    """
)


def embed_images(self, node, output_directory, render_prefix, source_path):
    paths_to_embed = []
    if node.get("pages"):
        for page_spec in node["pages"].split(","):
            page_spec = page_spec.strip()
            if "-" in page_spec:
                start_spec, _, stop_spec = page_spec.partition("-")
                start_page, stop_page = int(start_spec), int(stop_spec) + 1
            else:
                start_page, stop_page = int(page_spec), int(page_spec) + 1
            for page_number in range(start_page, stop_page):
                for path in output_directory.glob(f"{render_prefix}-{page_number}.svg"):
                    paths_to_embed.append(path)
    elif node.get("no-trim"):
        for path in output_directory.glob(f"{render_prefix}.svg"):
            paths_to_embed.append(path)
        if not paths_to_embed:
            page_count = len(list(output_directory.glob(f"{render_prefix}-*.svg")))
            for page_number in range(1, page_count + 1):
                for path in output_directory.glob(f"{render_prefix}-{page_number}.svg"):
                    paths_to_embed.append(path)
    else:
        for path in output_directory.glob(f"{render_prefix}.cropped.svg"):
            paths_to_embed.append(path)
    with_columns = node.get("with-columns")
    if with_columns:
        for i in range(0, len(paths_to_embed), with_columns):
            self.body.append(table_row_open_template)
            for path in paths_to_embed[i : i + with_columns]:
                relative_path = pathlib.Path(self.builder.imgpath) / path.name
                self.body.append(
                    thumbnail_template.format(
                        alt="",
                        cls="table-cell thumbnail",
                        group=f"group-{render_prefix}",
                        fullsize_path=relative_path,
                        thumbnail_path=relative_path,
                        title="",
                    )
                )
            self.body.append(table_row_close_template)
    else:
        for path in paths_to_embed:
            relative_path = pathlib.Path(self.builder.imgpath) / path.name
            self.body.append(
                basic_image_template.format(
                    relative_path=relative_path, source_path=source_path
                )
            )


def install_lightbox_static_files(app):
    source_static_path = os.path.join(app.builder.srcdir, "_static")
    target_static_path = os.path.join(app.builder.outdir, "_static")
    source_lightbox_path = os.path.join(source_static_path, "lightbox2")
    target_lightbox_path = os.path.join(target_static_path, "lightbox2")
    relative_file_paths = []
    for root, _, file_names in os.walk(source_lightbox_path):
        for file_name in file_names:
            absolute_file_path = os.path.join(root, file_name)
            relative_file_path = os.path.relpath(absolute_file_path, source_static_path)
            relative_file_paths.append(relative_file_path)
    if os.path.exists(target_lightbox_path):
        shutil.rmtree(target_lightbox_path)
    for relative_file_path in sphinx.util.display.status_iterator(
        relative_file_paths,
        "installing lightbox files... ",
        brown,
        len(relative_file_paths),
    ):
        source_path = os.path.join(source_static_path, relative_file_path)
        target_path = os.path.join(target_static_path, relative_file_path)
        target_directory = os.path.dirname(target_path)
        if not os.path.exists(target_directory):
            ensuredir(target_directory)
        copyfile(source_path, target_path)
        if relative_file_path.endswith(".js"):
            app.add_js_file(relative_file_path, defer="defer")
        elif relative_file_path.endswith(".css"):
            app.add_css_file(relative_file_path)


def on_html_collect_pages(app):
    for path in sphinx.util.display.status_iterator(
        app.env.thumbnails,
        "copying gallery thumbnails...",
        "brown",
        len(app.env.thumbnails),
    ):
        source_path = pathlib.Path(app.srcdir) / path
        target_path = pathlib.Path(app.outdir) / "_images" / source_path.name
        try:
            shutil.copy(source_path, target_path)
        except Exception:
            logger.warning(f"Could not copy {source_path}")
    return []


def setup(app):
    app.connect("builder-inited", on_builder_inited)
    app.connect("html-collect-pages", on_html_collect_pages)
    app.add_css_file("abjad.css")
    app.add_directive("docs", HiddenDoctestDirective)
    app.add_directive("shell", ShellDirective)
    app.add_directive("thumbnail", ThumbnailDirective)
    app.add_js_file("copybutton.js", defer="defer")
    app.add_js_file("ga.js")
    app.add_node(
        thumbnail_block,
        html=[visit_thumbnail_block_html, None],
        latex=[visit_thumbnail_block_latex, None],
    )



import collections
import dataclasses
import operator as operator_module
import typing

from . import math as _math


@dataclasses.dataclass(slots=True)
class Pattern:
    """
    Pattern.

    ..  container:: example

        Matches three indices out of every eight:

        >>> pattern = abjad.Pattern(
        ...     indices=[0, 1, 7],
        ...     period=8,
        ... )

        >>> total_length = 16
        >>> for index in range(16):
        ...     match = pattern.matches_index(index, total_length)
        ...     match = match or ''
        ...     print(index, match)
        0 True
        1 True
        2
        3
        4
        5
        6
        7 True
        8 True
        9 True
        10
        11
        12
        13
        14
        15 True

    ..  container:: example

        Matches three indices out of every sixteen:

        >>> pattern = abjad.Pattern(
        ...     indices=[0, 1, 7],
        ...     period=16,
        ... )

        >>> total_length = 16
        >>> for index in range(16):
        ...     match = pattern.matches_index(index, total_length)
        ...     match = match or ''
        ...     print(index, match)
        0 True
        1 True
        2
        3
        4
        5
        6
        7 True
        8
        9
        10
        11
        12
        13
        14
        15

    ..  container:: example

        Works with improper indices:

        >>> pattern = abjad.Pattern(
        ...     indices=[16, 17, 23],
        ...     period=16,
        ... )

        >>> total_length = 16
        >>> for index in range(16):
        ...     match = pattern.matches_index(index, total_length)
        ...     match = match or ''
        ...     print(index, match)
        0 True
        1 True
        2
        3
        4
        5
        6
        7 True
        8
        9
        10
        11
        12
        13
        14
        15

    ..  container:: example

        Sieve from opening of Xenakis's Psappha:

        >>> sieve_1a = abjad.index([0, 1, 7], 8)
        >>> sieve_1b = abjad.index([1, 3], 5)
        >>> sieve_1 = sieve_1a & sieve_1b
        >>> sieve_2a = abjad.index([0, 1, 2], 8)
        >>> sieve_2b = abjad.index([0], 5)
        >>> sieve_2 = sieve_2a & sieve_2b
        >>> sieve_3 = abjad.index([3], 8)
        >>> sieve_4 = abjad.index([4], 8)
        >>> sieve_5a = abjad.index([5, 6], 8)
        >>> sieve_5b = abjad.index([2, 3, 4], 5)
        >>> sieve_5 = sieve_5a & sieve_5b
        >>> sieve_6a = abjad.index([1], 8)
        >>> sieve_6b = abjad.index([2], 5)
        >>> sieve_6 = sieve_6a & sieve_6b
        >>> sieve_7a = abjad.index([6], 8)
        >>> sieve_7b = abjad.index([1], 5)
        >>> sieve_7 = sieve_7a & sieve_7b
        >>> sieve = sieve_1 | sieve_2 | sieve_3 | sieve_4 | sieve_5 | sieve_6 | sieve_7

        >>> sieve.get_boolean_vector(total_length=40)
        [1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1,
        1, 0, 1, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 0]

    ..  container:: example

        Inverted works like this:

        Matches three indices out of every eight:

        >>> pattern = abjad.Pattern(
        ...     indices=[0, 1, 7],
        ...     period=8,
        ... )

        >>> pattern.inverted is None
        True

        >>> total_length = 16
        >>> for index in range(16):
        ...     match = pattern.matches_index(index, total_length)
        ...     match = match or ''
        ...     print(index, match)
        0 True
        1 True
        2
        3
        4
        5
        6
        7 True
        8 True
        9 True
        10
        11
        12
        13
        14
        15 True

        Pattern that rejects three indices from every eight; equivalently, pattern
        matches ``8-3=5`` indices out of every eight:

        >>> pattern = abjad.Pattern(
        ...     indices=[0, 1, 7],
        ...     period=8,
        ...     inverted=True
        ... )

        >>> pattern.inverted
        True

        >>> total_length = 16
        >>> for index in range(16):
        ...     match = pattern.matches_index(index, total_length)
        ...     match = match or ''
        ...     print(index, match)
        0
        1
        2 True
        3 True
        4 True
        5 True
        6 True
        7
        8
        9
        10 True
        11 True
        12 True
        13 True
        14 True
        15

        Matches every index that is (one of the first three indices) OR (one of the last
        three indices):

        >>> pattern_1 = abjad.index_first(3)
        >>> pattern_2 = abjad.index_last(3)
        >>> pattern = pattern_1 | pattern_2
        >>> pattern.inverted is None
        True

        >>> pattern.get_boolean_vector(total_length=16)
        [1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1]

        Matches every index that is NOT (one of the first three indices) OR (one of the
        last three indices):

        >>> import dataclasses
        >>> pattern = dataclasses.replace(pattern, inverted=True)
        >>> pattern.inverted
        True

        >>> pattern.get_boolean_vector(total_length=16)
        [0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0]

    ..  container:: example

        Payload works like this: Pattern with string payload assigned to three of every
        eight indices:

        >>> pattern = abjad.Pattern(
        ...     indices=[0, 1, 7],
        ...     payload='Allegro non troppo',
        ...     period=8,
        ... )

        >>> total_length = 10
        >>> for index in range(10):
        ...     match = pattern.matches_index(index, total_length)
        ...     if match:
        ...         payload = pattern.payload
        ...     else:
        ...         payload = ''
        ...     print(index, repr(payload))
        ...
        0 'Allegro non troppo'
        1 'Allegro non troppo'
        2 ''
        3 ''
        4 ''
        5 ''
        6 ''
        7 'Allegro non troppo'
        8 'Allegro non troppo'
        9 'Allegro non troppo'

    ..  container:: example

        Period works like this:

        Pattern with a period of eight:

        >>> pattern = abjad.Pattern(
        ...     indices=[0, 1, 7],
        ...     period=8,
        ... )

        >>> pattern.period
        8

        >>> total_length = 16
        >>> for index in range(16):
        ...     match = pattern.matches_index(index, total_length)
        ...     match = match or ''
        ...     print(index, match)
        0 True
        1 True
        2
        3
        4
        5
        6
        7 True
        8 True
        9 True
        10
        11
        12
        13
        14
        15 True

        Same pattern with a period of sixteen:

        >>> pattern = abjad.Pattern(
        ...     indices=[0, 1, 7],
        ...     period=16,
        ... )

        >>> pattern.period
        16

        >>> total_length = 16
        >>> for index in range(16):
        ...     match = pattern.matches_index(index, total_length)
        ...     match = match or ''
        ...     print(index, match)
        0 True
        1 True
        2
        3
        4
        5
        6
        7 True
        8
        9
        10
        11
        12
        13
        14
        15

        Gets period of pattern that indexs every fourth and fifth element:

        >>> pattern_1 = abjad.Pattern([0], period=4)
        >>> pattern_2 = abjad.Pattern([0], period=5)
        >>> pattern = pattern_1 | pattern_2
        >>> pattern
        Pattern(indices=None, inverted=None, operator='or', patterns=(Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=4), Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=5)), payload=None, period=20)

        >>> pattern.period
        20

        Returns none when pattern contains acyclic parts:

        >>> pattern_1 = abjad.Pattern([0], period=4)
        >>> pattern_2 = abjad.Pattern([0])
        >>> pattern = pattern_1 | pattern_2
        >>> pattern
        Pattern(indices=None, inverted=None, operator='or', patterns=(Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=4), Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

        >>> pattern.period is None
        True

    """

    indices: typing.Any = None
    inverted: typing.Any = None
    operator: typing.Any = None
    patterns: typing.Any = None
    payload: typing.Any = None
    period: typing.Any = None

    _name_to_operator = {
        "and": operator_module.and_,
        "or": operator_module.or_,
        "xor": operator_module.xor,
    }

    def __post_init__(self):
        if self.indices is not None:
            assert all(isinstance(_, int) for _ in self.indices), repr(self.indices)
            self.indices = tuple(self.indices)
        if self.inverted is not None:
            self.inverted = bool(self.inverted)
        if self.operator is not None:
            assert self.operator in self._name_to_operator, repr(self.operator)
        if self.patterns is not None:
            assert all(isinstance(_, type(self)) for _ in self.patterns)
            self.patterns = tuple(self.patterns)
        if self.period is None:
            if self.patterns:
                periods = [_.period for _ in self.patterns]
                if None not in periods:
                    self.period = _math.least_common_multiple(*periods)
        if self.period is not None:
            assert _math.is_positive_integer(self.period), repr(self.period)

    ### SPECIAL METHODS ###

    def __and__(self, pattern):
        """
        Logical AND of two patterns.

        ..  container:: example

            Flat grouping of two patterns:

            >>> pattern_1 = abjad.index_first(3)
            >>> pattern_2 = abjad.index_last(3)
            >>> pattern_1 & pattern_2
            Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

        ..  container:: example

            Flat grouping of three patterns:

            >>> pattern_1 = abjad.index_first(3)
            >>> pattern_2 = abjad.index_last(3)
            >>> pattern_3 = abjad.index([0], 2)
            >>> pattern = pattern_1 & pattern_2 & pattern_3
            >>> pattern
            Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=2)), payload=None, period=None)

            >>> pattern.get_boolean_vector(total_length=16)
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        ..  container:: example

            Nested grouping of three patterns:

            >>> pattern_1 = abjad.index_first(3)
            >>> pattern_2 = abjad.index_last(3)
            >>> pattern_3 = abjad.index([0], 2)
            >>> pattern = pattern_1 & pattern_2 | pattern_3
            >>> pattern
            Pattern(indices=None, inverted=None, operator='or', patterns=(Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None), Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=2)), payload=None, period=None)

            >>> pattern.get_boolean_vector(total_length=16)
            [1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0]

        ..  container:: example

            In-place AND is allowed:

            >>> pattern = abjad.index_first(3)
            >>> pattern &= abjad.index_last(3)
            >>> pattern
            Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

        Returns new pattern.
        """
        if self._can_append_to_self(pattern, "and"):
            if self.patterns is None:
                self_patterns = [self]
            else:
                self_patterns = list(self.patterns)
            patterns = self_patterns + [pattern]
            result = type(self)(operator="and", patterns=patterns)
        else:
            result = type(self)(operator="and", patterns=[self, pattern])
        return result

    def __invert__(self):
        """
        Inverts pattern.

        ..  container:: example

            >>> pattern = abjad.index_first(3)
            >>> pattern
            Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None)

            >>> pattern = ~pattern
            >>> pattern
            Pattern(indices=(0, 1, 2), inverted=True, operator=None, patterns=None, payload=None, period=None)

            >>> pattern = ~pattern
            >>> pattern
            Pattern(indices=(0, 1, 2), inverted=False, operator=None, patterns=None, payload=None, period=None)

            Negation defined equal to inversion.

        ..  container:: example

            Matches every index that is (one of the first three indices) or (one of the
            last three indices):

            >>> pattern_1 = abjad.index_first(3)
            >>> pattern_2 = abjad.index_last(3)
            >>> pattern = pattern_1 | pattern_2
            >>> pattern
            Pattern(indices=None, inverted=None, operator='or', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

            >>> pattern.get_boolean_vector(total_length=16)
            [1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1]

        ..  container:: example

            Matches every index that is NOT (one of the first three indices) or
            (one of the last three indices):

            >>> pattern = ~pattern
            >>> pattern
            Pattern(indices=None, inverted=True, operator='or', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

            >>> pattern.get_boolean_vector(total_length=16)
            [0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0]

        Returns new pattern.
        """
        inverted = not self.inverted
        return dataclasses.replace(self, inverted=inverted)

    def __len__(self):
        """
        Gets length of pattern.

        ..  container:: example

            Gets length of cyclic pattern:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=8,
            ... )

            >>> len(pattern)
            8

            Length of cyclic pattern defined equal to period of the pattern.

        ..  container:: example

            Gets length of acyclic pattern:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 2, 3],
            ... )

            >>> len(pattern)
            4

            Length of acyclic pattern defined equal to greatest index in
            pattern, plus 1.

        ..  container:: example

            Gets length of pattern with negative indices:

            >>> pattern = abjad.Pattern(
            ...     indices=[-3],
            ... )

            >>> len(pattern)
            3

            Length of pattern with negative indices defined equal to absolute
            value of least index.

        Returns nonnegative integer.
        """
        if self.period is not None:
            return self.period
        if self.indices:
            absolute_indices = []
            for index in self.indices:
                if 0 <= index:
                    absolute_indices.append(index)
                else:
                    index = abs(index) - 1
                    absolute_indices.append(index)
            maximum_index = max(absolute_indices)
            return maximum_index + 1
        return 0

    def __or__(self, pattern):
        """
        Logical OR of two patterns.

        ..  container:: example

            Flat grouping of two patterns:

            >>> pattern_1 = abjad.index_first(3)
            >>> pattern_2 = abjad.index_last(3)
            >>> pattern = pattern_1 | pattern_2
            >>> pattern
            Pattern(indices=None, inverted=None, operator='or', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

        ..  container:: example

            Flat grouping:

            >>> pattern_1 = abjad.index_first(3)
            >>> pattern_2 = abjad.index_last(3)
            >>> pattern_3 = abjad.index([0], 2)
            >>> pattern = pattern_1 | pattern_2 | pattern_3
            >>> pattern
            Pattern(indices=None, inverted=None, operator='or', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=2)), payload=None, period=None)

            >>> pattern.get_boolean_vector(total_length=16)
            [1, 1, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1]

        ..  container:: example

            Nested grouping:

            >>> pattern_1 = abjad.index_first(3)
            >>> pattern_2 = abjad.index_last(3)
            >>> pattern_3 = abjad.index([0], 2)
            >>> pattern = pattern_1 | pattern_2 & pattern_3
            >>> pattern
            Pattern(indices=None, inverted=None, operator='or', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=2)), payload=None, period=None)), payload=None, period=None)

            >>> pattern.get_boolean_vector(total_length=16)
            [1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0]

        ..  container:: example

            In-place OR is allowed:

            >>> pattern = abjad.index_first(3)
            >>> pattern |= abjad.index_last(3)
            >>> pattern
            Pattern(indices=None, inverted=None, operator='or', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

        Returns new pattern.
        """
        if self._can_append_to_self(pattern, "or"):
            if self.patterns is None:
                self_patterns = [self]
            else:
                self_patterns = list(self.patterns)
            patterns = self_patterns + [pattern]
            result = type(self)(operator="or", patterns=patterns)
        else:
            result = type(self)(operator="or", patterns=[self, pattern])
        return result

    def __xor__(self, pattern):
        """
        Logical XOR of two patterns.

        ..  container:: example

            >>> pattern_1 = abjad.index_first(3)
            >>> pattern_2 = abjad.index_last(3)
            >>> pattern = pattern_1 ^ pattern_2
            >>> pattern
            Pattern(indices=None, inverted=None, operator='xor', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

        ..  container:: example

            Flat grouping:

            >>> pattern_1 = abjad.index_first(3)
            >>> pattern_2 = abjad.index_last(3)
            >>> pattern_3 = abjad.index([0], 2)
            >>> pattern = pattern_1 ^ pattern_2 ^ pattern_3
            >>> pattern
            Pattern(indices=None, inverted=None, operator='xor', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=2)), payload=None, period=None)

            >>> pattern.get_boolean_vector(total_length=16)
            [0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1]

        ..  container:: example

            Nested grouping:

            >>> pattern_1 = abjad.index_first(3)
            >>> pattern_2 = abjad.index_last(3)
            >>> pattern_3 = abjad.index([0], 2)
            >>> pattern = pattern_1 ^ pattern_2 & pattern_3
            >>> pattern
            Pattern(indices=None, inverted=None, operator='xor', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=2)), payload=None, period=None)), payload=None, period=None)

            >>> pattern.get_boolean_vector(total_length=16)
            [1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0]

        ..  container:: example

            In-place XOR is allowed:

            >>> pattern = abjad.index_first(3)
            >>> pattern ^= abjad.index_last(3)
            >>> pattern
            Pattern(indices=None, inverted=None, operator='xor', patterns=(Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

        Returns new pattern.
        """
        if self._can_append_to_self(pattern, "xor"):
            if self.patterns is None:
                self_patterns = [self]
            else:
                self_patterns = list(self.patterns)
            patterns = self_patterns + [pattern]
            result = type(self)(operator="xor", patterns=patterns)
        else:
            result = type(self)(operator="xor", patterns=[self, pattern])
        return result

    ### PRIVATE METHODS ###

    def _can_append_to_self(self, pattern, operator_):
        if not isinstance(pattern, type(self)):
            return False
        if self.operator is None:
            return True
        if self.operator == operator_ and (
            pattern.operator is None or (pattern.operator == self.operator)
        ):
            return True
        return False

    def _make_subscript_string(self):
        return str(self)

    @property
    def weight(self):
        """
        Gets weight of pattern.

        ..  container:: example

            Gets weight of cyclic pattern:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=8,
            ... )

            >>> pattern.weight
            3

        ..  container:: example

            Gets weight of acyclic pattern:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 2, 3],
            ... )

            >>> pattern.weight
            3

        Weight defined equal to number of indices in pattern.

        Returns nonnegative integer.
        """
        return len(self.indices)

    ### PUBLIC METHODS ###

    def advance(self, count: int | None = None) -> "Pattern":
        """
        Advances pattern.

        ..  container:: example

            >>> pattern = abjad.Pattern([0, 2, 12])

            >>> pattern = pattern.advance(8)
            >>> pattern
            Pattern(indices=(4,), inverted=None, operator=None, patterns=None, payload=None, period=None)

            >>> pattern = pattern.advance(8)
            >>> pattern
            Pattern(indices=(), inverted=None, operator=None, patterns=None, payload=None, period=None)

        ..  container:: example

            Returns copy of pattern when count is none:

            >>> pattern = abjad.Pattern([0, 2, 12])
            >>> pattern.advance()
            Pattern(indices=(0, 2, 12), inverted=None, operator=None, patterns=None, payload=None, period=None)

        ..  container:: example exception

            Raises exception on attempt to advance negative pattern:

            >>> pattern = abjad.Pattern([-2, -1])
            >>> pattern.advance(8)
            Traceback (most recent call last):
                ...
            Exception: can not advance pattern with negative indices ...

        """
        if not count:
            return dataclasses.replace(self)
        assert 0 < count, repr(count)
        for index in self.indices:
            if index < 0:
                message = "can not advance pattern"
                message += f" with negative indices ({repr(self)})."
                raise Exception(message)
        new_indices = []
        for index in self.indices:
            new_index = index - count
            if 0 <= new_index:
                new_indices.append(new_index)
        return dataclasses.replace(self, indices=new_indices)

    @classmethod
    def from_vector(class_, vector):
        """
        Makes pattern from boolean ``vector``.

        ..  container:: example

            Matches three indices out of every five:

            >>> pattern = [1, 0, 0, 1, 1]
            >>> pattern = abjad.Pattern.from_vector(pattern)
            >>> pattern
            Pattern(indices=(0, 3, 4), inverted=None, operator=None, patterns=None, payload=None, period=5)

            >>> total_length = 10
            >>> for index in range(10):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1
            2
            3 True
            4 True
            5 True
            6
            7
            8 True
            9 True

        ..  container:: example

            Matches three indices out of every six:

            >>> pattern = [1, 0, 0, 1, 1, 0]
            >>> pattern = abjad.Pattern.from_vector(pattern)
            >>> pattern
            Pattern(indices=(0, 3, 4), inverted=None, operator=None, patterns=None, payload=None, period=6)

            >>> total_length = 12
            >>> for index in range(12):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1
            2
            3 True
            4 True
            5
            6 True
            7
            8
            9 True
            10 True
            11

        Returns pattern.
        """
        vector = [bool(_) for _ in vector]
        period = len(vector)
        indices = [i for i, x in enumerate(vector) if x]
        return class_(period=period, indices=indices)

    def get_boolean_vector(self, total_length=None):
        """
        Gets boolean vector of pattern applied to input sequence with
        ``total_length``.

        ..  container:: example

            Gets boolean vector of acyclic pattern:

            >>> pattern = abjad.Pattern(
            ...     indices=[4, 5, 6, 7],
            ... )


            >>> pattern.get_boolean_vector(4)
            [0, 0, 0, 0]

            >>> pattern.get_boolean_vector(8)
            [0, 0, 0, 0, 1, 1, 1, 1]

            >>> pattern.get_boolean_vector(16)
            [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0]

            Sets total length to length of pattern when ``total_length`` is
            none:

            >>> pattern.get_boolean_vector()
            [0, 0, 0, 0, 1, 1, 1, 1]

        ..  container:: example

            Gets vector of cyclic pattern:

            >>> pattern = abjad.Pattern(
            ...     indices=[4, 5, 6, 7],
            ...     period=20,
            ... )

            >>> pattern.get_boolean_vector(4)
            [0, 0, 0, 0]

            >>> pattern.get_boolean_vector(8)
            [0, 0, 0, 0, 1, 1, 1, 1]

            >>> pattern.get_boolean_vector(16)
            [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0]

            Sets total length to length of pattern when ``total_length`` is
            none:

            >>> pattern.get_boolean_vector()
            [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        ..  container:: example

            Gets vector of inverted pattern:

            >>> pattern = abjad.Pattern(
            ...     indices=[4, 5, 6, 7],
            ...     period=20,
            ... )

            >>> pattern.get_boolean_vector(4)
            [0, 0, 0, 0]

            >>> pattern.get_boolean_vector(8)
            [0, 0, 0, 0, 1, 1, 1, 1]

            >>> pattern.get_boolean_vector(16)
            [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0]

            Sets total length to length of pattern when ``total_length`` is
            none:

            >>> pattern.get_boolean_vector()
            [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        ..  container:: example

            Two-part pattern with logical OR:

            >>> pattern = abjad.Pattern(
            ...     operator='or',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0, 1, 2],
            ...             ),
            ...         abjad.Pattern(
            ...             indices=[-3, -2, -1],
            ...             ),
            ...         ],
            ... )

            >>> pattern.get_boolean_vector(4)
            [1, 1, 1, 1]

            >>> pattern.get_boolean_vector(8)
            [1, 1, 1, 0, 0, 1, 1, 1]

            >>> pattern.get_boolean_vector(16)
            [1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1]

            Matches every index that is (one of the first three indices) OR
            (one of the last three indices).

        ..  container:: example

            Two-part pattern with mixed periodic and inverted parts:

            >>> pattern = abjad.Pattern(
            ...     operator='and',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0],
            ...             period=2,
            ...             ),
            ...         abjad.Pattern(
            ...             indices=[-3, -2, -1],
            ...             inverted=True,
            ...             ),
            ...         ],
            ... )

            >>> pattern.get_boolean_vector(4)
            [1, 0, 0, 0]

            >>> pattern.get_boolean_vector(8)
            [1, 0, 1, 0, 1, 0, 0, 0]

            >>> pattern.get_boolean_vector(16)
            [1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0]

            Matches every index that is (equal to 0 % 2) AND (not one of the
            last three indices).

        ..  container:: example

            Cyclic pattern that indexes every fourth and fifth item:

            >>> pattern_1 = abjad.Pattern([0], period=4)
            >>> pattern_2 = abjad.Pattern([0], period=5)
            >>> pattern = pattern_1 | pattern_2
            >>> pattern
            Pattern(indices=None, inverted=None, operator='or', patterns=(Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=4), Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=5)), payload=None, period=20)

            >>> pattern.get_boolean_vector(4)
            [1, 0, 0, 0]

            >>> pattern.get_boolean_vector(8)
            [1, 0, 0, 0, 1, 1, 0, 0]

            >>> pattern.get_boolean_vector(16)
            [1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0, 1, 0, 0, 1]

            Sets total length to period of pattern when ``total_length`` is
            none:

            >>> pattern.period
            20

            >>> pattern.get_boolean_vector()
            [1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 0]

            >>> pattern.period == len(pattern.get_boolean_vector())
            True

        Returns list of ones and zeroes.
        """
        total_length = total_length or len(self)
        boolean_vector = []
        for index in range(total_length):
            result = self.matches_index(index, total_length)
            boolean_vector.append(int(result))
        return boolean_vector

    def get_matching_items(self, sequence):
        """
        Gets maching items from sequence.

        ..  container:: example

            >>> pattern = abjad.Pattern(
            ...     indices=[4, 5, 6, 7],
            ... )

            >>> pattern.get_matching_items('abcdefghijklmnopqrstuvwxyz')
            ['e', 'f', 'g', 'h']

        ..  container:: example

            >>> pattern = abjad.Pattern(
            ...     indices=[8, 9],
            ...     period=10,
            ... )

            >>> pattern.get_matching_items('abcdefghijklmnopqrstuvwxyz')
            ['i', 'j', 's', 't']

        ..  container:: example

            >>> pattern = abjad.index_first(1) | abjad.index_last(2)

            >>> pattern.get_matching_items('abcdefghijklmnopqrstuvwxyz')
            ['a', 'y', 'z']

        Returns list.
        """
        assert isinstance(sequence, collections.abc.Iterable), repr(sequence)
        length = len(sequence)
        items = []
        for i in range(length):
            if self.matches_index(i, length):
                item = sequence[i]
                items.append(item)
        return items

    @staticmethod
    def index(indices, period=None, inverted=None):
        """
        Makes pattern that matches ``indices``.

        ..  container:: example

            Indexes item 2:

            >>> pattern = abjad.index([2])
            >>> pattern
            Pattern(indices=(2,), inverted=None, operator=None, patterns=None, payload=None, period=None)

        ..  container:: example

            Indexes items 2, 3 and 5:

            >>> pattern = abjad.index([2, 3, 5])
            >>> pattern
            Pattern(indices=(2, 3, 5), inverted=None, operator=None, patterns=None, payload=None, period=None)

        Returns pattern.
        """
        assert all(isinstance(_, int) for _ in indices), repr(indices)
        indices = indices or []
        return Pattern(
            indices=indices,
            inverted=inverted,
            period=period,
        )

    @staticmethod
    def index_all(inverted=None):
        """
        Makes pattern that matches all indices.

        ..  container:: example

            Indexes all divisions for tie creation:

            >>> pattern = abjad.index_all()
            >>> pattern
            Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=1)

        Returns pattern.
        """
        return Pattern(indices=[0], inverted=inverted, period=1)

    @staticmethod
    def index_first(n, inverted=None):
        """
        Makes pattern that matches the first ``n`` indices.

        ..  container:: example

            Indexes first item:

            >>> pattern = abjad.index_first(1)
            >>> pattern
            Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=None)

        ..  container:: example

            Indexes first two items:

            >>> pattern = abjad.index_first(2)
            >>> pattern
            Pattern(indices=(0, 1), inverted=None, operator=None, patterns=None, payload=None, period=None)

        ..  container:: example

            Indexes nothing:

            >>> pattern = abjad.index_first(0)
            >>> pattern
            Pattern(indices=None, inverted=None, operator=None, patterns=None, payload=None, period=None)

        Returns pattern.
        """
        assert isinstance(n, int), repr(n)
        if 0 < n:
            indices = list(range(n))
        else:
            indices = None
        return Pattern(indices=indices, inverted=inverted)

    @staticmethod
    def index_last(n, inverted=None):
        """
        Makes pattern that matches the last ``n`` indices.

        ..  container:: example

            Indexes last two items:

            >>> pattern = abjad.index_last(2)
            >>> pattern
            Pattern(indices=(-2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)

        ..  container:: example

            Indexes nothing:

            >>> pattern = abjad.index_last(0)
            >>> pattern
            Pattern(indices=None, inverted=None, operator=None, patterns=None, payload=None, period=None)

        Returns pattern.
        """
        assert isinstance(n, int), repr(n)
        if 0 < n:
            start = -1
            stop = -n - 1
            stride = -1
            indices = list(reversed(range(start, stop, stride)))
        else:
            indices = None
        return Pattern(indices=indices, inverted=inverted)

    def matches_index(self, index, total_length, rotation=None):
        """
        Is true when pattern matches ``index`` taken under ``total_length``.

        ..  container:: example

            Matches three indices out of every eight:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=8,
            ... )

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2
            3
            4
            5
            6
            7 True
            8 True
            9 True
            10
            11
            12
            13
            14
            15 True

            Matches three indices out of every eight, offset ``1`` to the left:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=8,
            ... )

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(
            ...         index,
            ...         total_length,
            ...         rotation=1,
            ...         )
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1
            2
            3
            4
            5
            6 True
            7 True
            8 True
            9
            10
            11
            12
            13
            14 True
            15 True

            Matches three indices out of every eight, offset ``2`` to the
            left:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=8,
            ... )

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(
            ...         index,
            ...         total_length,
            ...         rotation=2,
            ...         )
            ...     match = match or ''
            ...     print(index, match)
            0
            1
            2
            3
            4
            5 True
            6 True
            7 True
            8
            9
            10
            11
            12
            13 True
            14 True
            15 True

        ..  container:: example

            Matches three indices out of every sixteen:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=16,
            ... )

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2
            3
            4
            5
            6
            7 True
            8
            9
            10
            11
            12
            13
            14
            15

            Matches three indices out of every sixteen, offset ``1`` to the
            left:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=16,
            ... )

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(
            ...         index,
            ...         total_length,
            ...         rotation=1,
            ...         )
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1
            2
            3
            4
            5
            6 True
            7
            8
            9
            10
            11
            12
            13
            14
            15 True

            Matches three indices out of every sixteen, offset ``2`` to the
            left:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=16,
            ... )

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(
            ...         index,
            ...         total_length,
            ...         rotation=2,
            ...         )
            ...     match = match or ''
            ...     print(index, match)
            0
            1
            2
            3
            4
            5 True
            6
            7
            8
            9
            10
            11
            12
            13
            14 True
            15 True

        ..  container:: example

            Empty pattern:

            >>> pattern = abjad.Pattern()

            Total length 16:

            >>> total_length = 16
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0
            1
            2
            3
            4
            5
            6
            7
            8
            9
            10
            11
            12
            13
            14
            15

            Total length 8:

            >>> total_length = 8
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0
            1
            2
            3
            4
            5
            6
            7

            Total length 4:

            >>> total_length = 4
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0
            1
            2
            3

            Matches nothing.

        ..  container:: example

            Simple pattern:

            Logical OR:

            >>> pattern = abjad.Pattern(
            ...     operator='or',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0, 1, 2],
            ...             ),
            ...         ],
            ... )
            >>> total_length = 16
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3
            4
            5
            6
            7
            8
            9
            10
            11
            12
            13
            14
            15

            Logical AND:

            >>> pattern = abjad.Pattern(
            ...     operator='and',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0, 1, 2],
            ...             ),
            ...         ],
            ... )
            >>> total_length = 16
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3
            4
            5
            6
            7
            8
            9
            10
            11
            12
            13
            14
            15

            Logical XOR:

            >>> pattern = abjad.Pattern(
            ...     operator='xor',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0, 1, 2],
            ...             ),
            ...         ],
            ... )
            >>> total_length = 16
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3
            4
            5
            6
            7
            8
            9
            10
            11
            12
            13
            14
            15

            Matches every index that is (one of the first three indices).

            Ignores ``operator``.

        ..  container:: example

            Two-part pattern with logical OR:

            >>> pattern = abjad.Pattern(
            ...     operator='or',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0, 1, 2],
            ...             ),
            ...         abjad.Pattern(
            ...             indices=[-3, -2, -1],
            ...             ),
            ...         ],
            ... )

            Total length 16:

            >>> total_length = 16
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3
            4
            5
            6
            7
            8
            9
            10
            11
            12
            13 True
            14 True
            15 True

            Total length 8:

            >>> total_length = 8
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3
            4
            5 True
            6 True
            7 True

            Total length 4:

            >>> total_length = 4
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3 True

            Matches every index that is (one of the first three indices) OR
            (one of the last three indices).

        ..  container:: example

            Two-part pattern with logical AND:

            >>> pattern = abjad.Pattern(
            ...     operator='and',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0, 1, 2],
            ...             ),
            ...         abjad.Pattern(
            ...             indices=[-3, -2, -1],
            ...             ),
            ...         ],
            ... )

            Total length 16:

            >>> total_length = 16
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0
            1
            2
            3
            4
            5
            6
            7
            8
            9
            10
            11
            12
            13
            14
            15

            Total length 8:

            >>> total_length = 8
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0
            1
            2
            3
            4
            5
            6
            7

            Total length 4:

            >>> total_length = 4
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0
            1 True
            2 True
            3

            Matches every index that is (one of the first three indices) AND
            (one of the last three indices).

        ..  container:: example

            Two-part pattern with logical XOR:

            >>> pattern = abjad.Pattern(
            ...     operator='xor',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0, 1, 2],
            ...             ),
            ...         abjad.Pattern(
            ...             indices=[-3, -2, -1],
            ...             ),
            ...         ],
            ... )

            Total length 16:

            >>> total_length = 16
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3
            4
            5
            6
            7
            8
            9
            10
            11
            12
            13 True
            14 True
            15 True

            Total length 8:

            >>> total_length = 8
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3
            4
            5 True
            6 True
            7 True

            Total length 4:

            >>> total_length = 4
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1
            2
            3 True

            Matches every index that is (one of the first three indices) XOR
            (one of the last three indices).

        ..  container:: example

            Two-part pattern with mixed periodic and inverted parts:

            >>> pattern = abjad.Pattern(
            ...     operator='and',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0],
            ...             period=2,
            ...             ),
            ...         abjad.Pattern(
            ...             indices=[-3, -2, -1],
            ...             inverted=True,
            ...             ),
            ...         ],
            ... )

            Total length 16:

            >>> total_length = 16
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1
            2 True
            3
            4 True
            5
            6 True
            7
            8 True
            9
            10 True
            11
            12 True
            13
            14
            15

            Total length 8:

            >>> total_length = 8
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1
            2 True
            3
            4 True
            5
            6
            7

            Total length 4:

            >>> total_length = 4
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1
            2
            3

            Matches every index that is (equal to 0 % 2) AND (not one of the
            last three indices).

        ..  container:: example

            Complex pattern with compound and simple parts:

            >>> pattern = abjad.Pattern(
            ...     operator='and',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0],
            ...             period=2,
            ...             ),
            ...         abjad.Pattern(
            ...             indices=[-3, -2, -1],
            ...             inverted=True,
            ...             ),
            ...         ],
            ... )
            >>> pattern = abjad.Pattern(
            ...     operator='or',
            ...     patterns=[
            ...         pattern,
            ...         abjad.Pattern(
            ...             indices=[0, 1, 2],
            ...             ),
            ...         ],
            ... )
            >>> pattern
            Pattern(indices=None, inverted=None, operator='or', patterns=(Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=2), Pattern(indices=(-3, -2, -1), inverted=True, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None), Pattern(indices=(0, 1, 2), inverted=None, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

            Total length 16:

            >>> total_length = 16
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3
            4 True
            5
            6 True
            7
            8 True
            9
            10 True
            11
            12 True
            13
            14
            15

            Total length 8:

            >>> total_length = 8
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3
            4 True
            5
            6
            7

            Total length 4:

            >>> total_length = 4
            >>> for index in range(total_length):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2 True
            3

            Matches every index that is ((equal to 0 % 2) AND (not one of the
            last three indices)) OR is (one of the first three indices).

        Returns true or false.
        """
        if not self.patterns:
            assert 0 <= total_length
            if 0 <= index:
                nonnegative_index = index
            else:
                nonnegative_index = total_length - abs(index)
            inverted = bool(self.inverted)
            if not self.indices:
                return False ^ inverted
            if self.period is None:
                for index in self.indices:
                    if index < 0:
                        index = total_length - abs(index)
                    if index == nonnegative_index and index < total_length:
                        return True ^ inverted
            else:
                if rotation is not None:
                    nonnegative_index += rotation
                nonnegative_index = nonnegative_index % self.period
                for index in self.indices:
                    if index < 0:
                        index = total_length - abs(index)
                        index = index % self.period
                    if index == nonnegative_index and index < total_length:
                        return True ^ inverted
                    if (index % self.period) == nonnegative_index and (
                        index % self.period < total_length
                    ):
                        return True ^ inverted
            return False ^ inverted
        elif len(self.patterns) == 1:
            pattern = self.patterns[0]
            result = pattern.matches_index(index, total_length, rotation=rotation)
        else:
            operator_ = self._name_to_operator[self.operator]
            pattern = self.patterns[0]
            result = pattern.matches_index(index, total_length, rotation=rotation)
            for pattern in self.patterns[1:]:
                result_ = pattern.matches_index(index, total_length, rotation=rotation)
                result = operator_(result, result_)
        if self.inverted:
            result = not (result)
        return result

    def reverse(self):
        """
        Reverses pattern.

        ..  container:: example

            Matches three indices out of every eight:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=8,
            ... )

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2
            3
            4
            5
            6
            7 True
            8 True
            9 True
            10
            11
            12
            13
            14
            15 True

        ..  container:: example

            Reverses pattern:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=8,
            ... )

            >>> pattern = pattern.reverse()
            >>> pattern
            Pattern(indices=(-1, -2, -8), inverted=None, operator=None, patterns=None, payload=None, period=8)

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1
            2
            3
            4
            5
            6 True
            7 True
            8 True
            9
            10
            11
            12
            13
            14 True
            15 True

        ..  container:: example

            Matches every index that is (equal to 0 % 2) AND (not one of the
            last three indices):

            >>> pattern = abjad.Pattern(
            ...     operator='and',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0],
            ...             period=2,
            ...             ),
            ...         abjad.Pattern(
            ...             indices=[-3, -2, -1],
            ...             inverted=True,
            ...             ),
            ...         ],
            ... )
            >>> pattern
            Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=2), Pattern(indices=(-3, -2, -1), inverted=True, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

            Reverses pattern:

            >>> pattern = pattern.reverse()
            >>> pattern
            Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(-1,), inverted=None, operator=None, patterns=None, payload=None, period=2), Pattern(indices=(2, 1, 0), inverted=True, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

            New pattern matches every index that is (equal to -1 % 2) AND (not one of the
            first three indices).

        Returns new pattern.
        """
        if not self.patterns:
            indices = [-index - 1 for index in self.indices]
            return dataclasses.replace(self, indices=indices)
        patterns = [_.reverse() for _ in self.patterns]
        return dataclasses.replace(self, patterns=patterns)

    def rotate(self, n=0):
        """
        Rotates pattern by index ``n``.

        ..  container:: example

            Matches three indices out of every eight:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=8,
            ... )

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2
            3
            4
            5
            6
            7 True
            8 True
            9 True
            10
            11
            12
            13
            14
            15 True

            Rotates pattern two elements to the right:

            >>> pattern = abjad.Pattern(
            ...     indices=[0, 1, 7],
            ...     period=8,
            ... )

            >>> pattern = pattern.rotate(n=2)
            >>> pattern
            Pattern(indices=(2, 3, 9), inverted=None, operator=None, patterns=None, payload=None, period=8)

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0
            1 True
            2 True
            3 True
            4
            5
            6
            7
            8
            9 True
            10 True
            11 True
            12
            13
            14
            15

        ..  container:: example

            Matches three indices out of every eight with negative indices:

            >>> pattern = abjad.Pattern(
            ...     indices=[-3, -2, -1],
            ...     period=8,
            ... )

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0
            1
            2
            3
            4
            5 True
            6 True
            7 True
            8
            9
            10
            11
            12
            13 True
            14 True
            15 True

            Rotates pattern two elements to the right:

            >>> pattern = abjad.Pattern(
            ...     indices=[-3, -2, -1],
            ...     period=8,
            ... )

            >>> pattern = pattern.rotate(n=2)
            >>> pattern
            Pattern(indices=(-1, 0, 1), inverted=None, operator=None, patterns=None, payload=None, period=8)

            >>> total_length = 16
            >>> for index in range(16):
            ...     match = pattern.matches_index(index, total_length)
            ...     match = match or ''
            ...     print(index, match)
            0 True
            1 True
            2
            3
            4
            5
            6
            7 True
            8 True
            9 True
            10
            11
            12
            13
            14
            15 True

        ..  container:: example

            Matches every index that is (equal to 0 % 2) AND (not one of the
            last three indices):

            >>> pattern = abjad.Pattern(
            ...     operator='and',
            ...     patterns=[
            ...         abjad.Pattern(
            ...             indices=[0],
            ...             period=2,
            ...             ),
            ...         abjad.Pattern(
            ...             indices=[-3, -2, -1],
            ...             inverted=True,
            ...             ),
            ...         ],
            ... )
            >>> pattern
            Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(0,), inverted=None, operator=None, patterns=None, payload=None, period=2), Pattern(indices=(-3, -2, -1), inverted=True, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

            Rotates pattern two elements to the right:

            >>> pattern = pattern.rotate(n=2)
            >>> pattern
            Pattern(indices=None, inverted=None, operator='and', patterns=(Pattern(indices=(2,), inverted=None, operator=None, patterns=None, payload=None, period=2), Pattern(indices=(-1, 0, 1), inverted=True, operator=None, patterns=None, payload=None, period=None)), payload=None, period=None)

            New pattern matches every index that is (equal to 2 % 2) AND (not the first,
            second or last index in the pattern).

        Returns new pattern.
        """
        if not self.patterns:
            indices = [index + n for index in self.indices]
            return dataclasses.replace(self, indices=indices)
        patterns = [_.rotate(n=n) for _ in self.patterns]
        return dataclasses.replace(self, patterns=patterns)


@dataclasses.dataclass(slots=True, unsafe_hash=True)
class PatternTuple:
    """
    Pattern tuple.

    ..  container:: example

        Three patterns:

        >>> patterns = abjad.PatternTuple([
        ...     abjad.Pattern(
        ...         indices=[0, 1, 7],
        ...         period=10,
        ...         ),
        ...     abjad.Pattern(
        ...         indices=[-2, -1],
        ...         ),
        ...     abjad.Pattern(
        ...         indices=[2],
        ...         period=3,
        ...         ),
        ...     ])
        >>> patterns
        PatternTuple(items=(Pattern(indices=(0, 1, 7), inverted=None, operator=None, patterns=None, payload=None, period=10), Pattern(indices=(-2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None), Pattern(indices=(2,), inverted=None, operator=None, patterns=None, payload=None, period=3)))

    ..  container:: example

        Two patterns:

        >>> patterns = abjad.PatternTuple([
        ...     abjad.Pattern(
        ...         indices=[1],
        ...         period=2,
        ...         ),
        ...     abjad.Pattern(
        ...         indices=[-3, -2, -1],
        ...         ),
        ...     ])
        >>> patterns
        PatternTuple(items=(Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2), Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)))

    """

    items: typing.Sequence = ()

    def __post_init__(self):
        self.items = tuple(self.items or [])

    def get_matching_pattern(self, index, total_length, rotation=None):
        """
        Gets pattern matching ``index``.

        ..  container:: example

            Two patterns:

            >>> patterns = abjad.PatternTuple([
            ...     abjad.Pattern(
            ...         indices=[1],
            ...         period=2,
            ...         ),
            ...     abjad.Pattern(
            ...         indices=[-3, -2, -1],
            ...         ),
            ...     ])

            Gets patterns that match the first ten indices:

            >>> for i in range(10):
            ...     match = patterns.get_matching_pattern(i, 10)
            ...     print(i, match)
            ...
            0 None
            1 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            2 None
            3 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            4 None
            5 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            6 None
            7 Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)
            8 Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)
            9 Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)

            Last three indices match the second pattern.

            Gets patterns that match next ten indices:

            >>> for i in range(10, 20):
            ...     match = patterns.get_matching_pattern(i, 10)
            ...     print(i, match)
            ...
            10 None
            11 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            12 None
            13 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            14 None
            15 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            16 None
            17 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            18 None
            19 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)

            Last three indices no longer match the second pattern.

        ..  container:: example

            Gets patterns that match the first ten indices, with rotation set
            to ``1``:

            >>> for i in range(10):
            ...     match = patterns.get_matching_pattern(i, 10, rotation=1)
            ...     print(i, match)
            ...
            0 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            1 None
            2 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            3 None
            4 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            5 None
            6 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            7 Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)
            8 Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)
            9 Pattern(indices=(-3, -2, -1), inverted=None, operator=None, patterns=None, payload=None, period=None)

            Matching indices of first pattern offset by ``1``.

            Gets patterns that match next ten indices with rotation set to
            ``1``:

            >>> for i in range(10, 20):
            ...     match = patterns.get_matching_pattern(i, 10, rotation=1)
            ...     print(i, match)
            ...
            10 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            11 None
            12 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            13 None
            14 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            15 None
            16 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            17 None
            18 Pattern(indices=(1,), inverted=None, operator=None, patterns=None, payload=None, period=2)
            19 None

            Matching indices of first pattern offset by ``1``.

        ..  container:: example

            With inverted patterns:

            >>> patterns = abjad.PatternTuple([
            ...     abjad.Pattern(
            ...         indices=[-3],
            ...         inverted=True,
            ...         ),
            ...     ])

            >>> for i in range(10):
            ...     match = patterns.get_matching_pattern(i, 10)
            ...     print(i, match)
            ...
            0 Pattern(indices=(-3,), inverted=True, operator=None, patterns=None, payload=None, period=None)
            1 Pattern(indices=(-3,), inverted=True, operator=None, patterns=None, payload=None, period=None)
            2 Pattern(indices=(-3,), inverted=True, operator=None, patterns=None, payload=None, period=None)
            3 Pattern(indices=(-3,), inverted=True, operator=None, patterns=None, payload=None, period=None)
            4 Pattern(indices=(-3,), inverted=True, operator=None, patterns=None, payload=None, period=None)
            5 Pattern(indices=(-3,), inverted=True, operator=None, patterns=None, payload=None, period=None)
            6 Pattern(indices=(-3,), inverted=True, operator=None, patterns=None, payload=None, period=None)
            7 None
            8 Pattern(indices=(-3,), inverted=True, operator=None, patterns=None, payload=None, period=None)
            9 Pattern(indices=(-3,), inverted=True, operator=None, patterns=None, payload=None, period=None)

        Returns pattern or none.
        """
        for pattern in reversed(self.items):
            if hasattr(pattern, "pattern"):
                if pattern.pattern.matches_index(
                    index, total_length, rotation=rotation
                ):
                    return pattern
            elif pattern.matches_index(index, total_length, rotation=rotation):
                return pattern

    def get_matching_payload(self, index, total_length, rotation=None):
        """
        Gets payload attached to pattern matching ``index``.

        ..  container:: example

            Two patterns. Underlying notes with even divisions
            assigned to the last three indices:

            >>> patterns = abjad.PatternTuple([
            ...     abjad.Pattern(
            ...         indices=[0],
            ...         payload='staccato',
            ...         period=1,
            ...         ),
            ...     abjad.Pattern(
            ...         indices=[-3, -2, -1],
            ...         payload='tenuto',
            ...         ),
            ...     ])

            Over ten indices:

            >>> for i in range(10):
            ...     match = patterns.get_matching_payload(i, 10)
            ...     print(i, match)
            ...
            0 staccato
            1 staccato
            2 staccato
            3 staccato
            4 staccato
            5 staccato
            6 staccato
            7 tenuto
            8 tenuto
            9 tenuto

            Over fifteen indices:

            >>> for i in range(15):
            ...     match = patterns.get_matching_payload(i, 15)
            ...     print(i, match)
            ...
            0 staccato
            1 staccato
            2 staccato
            3 staccato
            4 staccato
            5 staccato
            6 staccato
            7 staccato
            8 staccato
            9 staccato
            10 staccato
            11 staccato
            12 tenuto
            13 tenuto
            14 tenuto

        """
        pattern = self.get_matching_pattern(index, total_length, rotation=rotation)
        payload = None
        if pattern:
            payload = pattern.payload
        return payload



import collections
import fractions

from . import math as _math
from . import score as _score


class Parentage(collections.abc.Sequence):
    r'''
    Parentage of a component.

    ..  container:: example

        >>> score = abjad.Score()
        >>> string = r"""\new Voice = "Treble_Voice" { e'4 }"""
        >>> treble_staff = abjad.Staff(string, name="Treble_Staff")
        >>> score.append(treble_staff)
        >>> string = r"""\new Voice = "Bass_Voice" { c4 }"""
        >>> bass_staff = abjad.Staff(string, name="Bass_Staff")
        >>> clef = abjad.Clef('bass')
        >>> abjad.attach(clef, bass_staff[0][0])
        >>> score.append(bass_staff)
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \context Staff = "Treble_Staff"
                {
                    \context Voice = "Treble_Voice"
                    {
                        e'4
                    }
                }
                \context Staff = "Bass_Staff"
                {
                    \context Voice = "Bass_Voice"
                    {
                        \clef "bass"
                        c4
                    }
                }
            >>

        >>> bass_voice = score["Bass_Voice"]
        >>> note = bass_voice[0]
        >>> for component in abjad.get.parentage(note):
        ...     component
        ...
        Note('c4')
        Voice('c4', name='Bass_Voice')
        Staff('{ c4 }', name='Bass_Staff')
        Score("{ { e'4 } } { { c4 } }", simultaneous=True)

    '''

    ### CLASS VARIABLES ###

    __slots__ = ("_component", "_components")

    ### INITIALIZER ###

    def __init__(self, component=None):
        components = []
        if component is not None:
            assert isinstance(component, _score.Component), repr(component)
            components.extend(component._get_parentage())
        self._component = component
        self._components = tuple(components)

    ### SPECIAL METHODS ###

    def __eq__(self, argument):
        """
        Is true when ``argument`` is a parent with the same components as this parentage.
        """
        if isinstance(argument, type(self)):
            if len(self) == len(argument):
                for c, d in zip(self, argument):
                    if c is not d:
                        return False
                else:
                    return True
        return False

    def __getitem__(self, argument):
        """
        Gets ``argument``.

        Returns component or tuple of components.
        """
        return self.components.__getitem__(argument)

    def __len__(self) -> int:
        """
        Gets number of components in parentage.
        """
        return len(self.components)

    def __repr__(self):
        """
        Gets repr.
        """
        return f"{type(self).__name__}(component={self.component!r})"

    ### PRIVATE METHODS ###

    @staticmethod
    def _id_string(component):
        lhs = component.__class__.__name__
        rhs = getattr(component, "name", None) or id(component)
        return f"{lhs}-{rhs!r}"

    def _prolations(self):
        prolations = []
        default = fractions.Fraction(1)
        for parent in self:
            prolation = getattr(parent, "implied_prolation", default)
            prolations.append(prolation)
        return prolations

    ### PUBLIC PROPERTIES ###

    @property
    def component(self) -> _score.Component:
        r"""
        Gets component.

        ..  container:: example

            REGRESSION. Works with grace notes (and containers):

            >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
            >>> container = abjad.BeforeGraceContainer("cs'16")
            >>> abjad.attach(container, music_voice[1])
            >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
            >>> abjad.attach(abjad.Articulation(">"), obgc[0])
            >>> container = abjad.AfterGraceContainer("fs'16")
            >>> abjad.attach(container, music_voice[3])
            >>> staff = abjad.Staff([music_voice])
            >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \context Voice = "MusicVoice"
                    {
                        c'4
                        \grace {
                            cs'16
                        }
                        d'4
                        <<
                            \context Voice = "On_Beat_Grace_Container"
                            {
                                \set fontSize = #-3
                                \slash
                                \voiceOne
                                <
                                    \tweak font-size 0
                                    \tweak transparent ##t
                                    e'
                                    g'
                                >16
                                - \accent
                                [
                                (
                                gs'16
                                a'16
                                as'16
                                )
                                ]
                            }
                            \context Voice = "MusicVoice"
                            {
                                \voiceTwo
                                e'4
                            }
                        >>
                        \oneVoice
                        \afterGrace
                        f'4
                        {
                            fs'16
                        }
                    }
                }

            >>> for component in abjad.select.components(staff):
            ...     parentage = abjad.get.parentage(component)
            ...     print(f"{repr(component):30} {repr(parentage.component)}")
            Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }") Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice') Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
            Note("c'4")                    Note("c'4")
            BeforeGraceContainer("cs'16")  BeforeGraceContainer("cs'16")
            Note("cs'16")                  Note("cs'16")
            Note("d'4")                    Note("d'4")
            Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }") Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
            OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16") OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
            Chord("<e' g'>16")             Chord("<e' g'>16")
            Note("gs'16")                  Note("gs'16")
            Note("a'16")                   Note("a'16")
            Note("as'16")                  Note("as'16")
            Voice("e'4", name='MusicVoice') Voice("e'4", name='MusicVoice')
            Note("e'4")                    Note("e'4")
            Note("f'4")                    Note("f'4")
            AfterGraceContainer("fs'16")   AfterGraceContainer("fs'16")
            Note("fs'16")                  Note("fs'16")

        """
        return self._component

    @property
    def components(self) -> tuple[_score.Component]:
        r"""
        Gets components.

        ..  container:: example

            REGRESSION. Works with grace notes (and containers):

            >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
            >>> container = abjad.BeforeGraceContainer("cs'16")
            >>> abjad.attach(container, music_voice[1])
            >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
            >>> abjad.attach(abjad.Articulation(">"), obgc[0])
            >>> container = abjad.AfterGraceContainer("fs'16")
            >>> abjad.attach(container, music_voice[3])
            >>> staff = abjad.Staff([music_voice])
            >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \context Voice = "MusicVoice"
                    {
                        c'4
                        \grace {
                            cs'16
                        }
                        d'4
                        <<
                            \context Voice = "On_Beat_Grace_Container"
                            {
                                \set fontSize = #-3
                                \slash
                                \voiceOne
                                <
                                    \tweak font-size 0
                                    \tweak transparent ##t
                                    e'
                                    g'
                                >16
                                - \accent
                                [
                                (
                                gs'16
                                a'16
                                as'16
                                )
                                ]
                            }
                            \context Voice = "MusicVoice"
                            {
                                \voiceTwo
                                e'4
                            }
                        >>
                        \oneVoice
                        \afterGrace
                        f'4
                        {
                            fs'16
                        }
                    }
                }

            >>> for component in abjad.select.components(staff):
            ...     parentage = abjad.get.parentage(component)
            ...     components = parentage.components
            ...     print(f"{repr(component)}:")
            ...     for component_ in components:
            ...         print(f"    {repr(component_)}")
            Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }"):
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice'):
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("c'4"):
                Note("c'4")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            BeforeGraceContainer("cs'16"):
                BeforeGraceContainer("cs'16")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("cs'16"):
                Note("cs'16")
                BeforeGraceContainer("cs'16")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("d'4"):
                Note("d'4")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }"):
                Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16"):
                OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
                Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Chord("<e' g'>16"):
                Chord("<e' g'>16")
                OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
                Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("gs'16"):
                Note("gs'16")
                OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
                Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("a'16"):
                Note("a'16")
                OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
                Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("as'16"):
                Note("as'16")
                OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
                Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Voice("e'4", name='MusicVoice'):
                Voice("e'4", name='MusicVoice')
                Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("e'4"):
                Note("e'4")
                Voice("e'4", name='MusicVoice')
                Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("f'4"):
                Note("f'4")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            AfterGraceContainer("fs'16"):
                AfterGraceContainer("fs'16")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("fs'16"):
                Note("fs'16")
                AfterGraceContainer("fs'16")
                Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
                Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")

        """
        return self._components

    @property
    def orphan(self) -> bool:
        r"""
        Is true when component has no parent.

        ..  container:: example

            REGRESSION. Works with grace notes (and containers):

            >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
            >>> container = abjad.BeforeGraceContainer("cs'16")
            >>> abjad.attach(container, music_voice[1])
            >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
            >>> abjad.attach(abjad.Articulation(">"), obgc[0])
            >>> container = abjad.AfterGraceContainer("fs'16")
            >>> abjad.attach(container, music_voice[3])
            >>> staff = abjad.Staff([music_voice])
            >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \context Voice = "MusicVoice"
                    {
                        c'4
                        \grace {
                            cs'16
                        }
                        d'4
                        <<
                            \context Voice = "On_Beat_Grace_Container"
                            {
                                \set fontSize = #-3
                                \slash
                                \voiceOne
                                <
                                    \tweak font-size 0
                                    \tweak transparent ##t
                                    e'
                                    g'
                                >16
                                - \accent
                                [
                                (
                                gs'16
                                a'16
                                as'16
                                )
                                ]
                            }
                            \context Voice = "MusicVoice"
                            {
                                \voiceTwo
                                e'4
                            }
                        >>
                        \oneVoice
                        \afterGrace
                        f'4
                        {
                            fs'16
                        }
                    }
                }

            >>> for component in abjad.select.components(staff):
            ...     parentage = abjad.get.parentage(component)
            ...     print(f"{repr(component):30} {repr(parentage.orphan)}")
            Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }") True
            Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice') False
            Note("c'4")                    False
            BeforeGraceContainer("cs'16")  False
            Note("cs'16")                  False
            Note("d'4")                    False
            Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }") False
            OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16") False
            Chord("<e' g'>16")             False
            Note("gs'16")                  False
            Note("a'16")                   False
            Note("as'16")                  False
            Voice("e'4", name='MusicVoice') False
            Note("e'4")                    False
            Note("f'4")                    False
            AfterGraceContainer("fs'16")   False
            Note("fs'16")                  False

        """
        return self.parent is None

    @property
    def parent(self) -> _score.Component | None:
        r"""
        Gets parent.

        ..  container:: example

            REGRESSION. Works with grace notes (and containers):

            >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
            >>> container = abjad.BeforeGraceContainer("cs'16")
            >>> abjad.attach(container, music_voice[1])
            >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
            >>> abjad.attach(abjad.Articulation(">"), obgc[0])
            >>> container = abjad.AfterGraceContainer("fs'16")
            >>> abjad.attach(container, music_voice[3])
            >>> staff = abjad.Staff([music_voice])
            >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \context Voice = "MusicVoice"
                    {
                        c'4
                        \grace {
                            cs'16
                        }
                        d'4
                        <<
                            \context Voice = "On_Beat_Grace_Container"
                            {
                                \set fontSize = #-3
                                \slash
                                \voiceOne
                                <
                                    \tweak font-size 0
                                    \tweak transparent ##t
                                    e'
                                    g'
                                >16
                                - \accent
                                [
                                (
                                gs'16
                                a'16
                                as'16
                                )
                                ]
                            }
                            \context Voice = "MusicVoice"
                            {
                                \voiceTwo
                                e'4
                            }
                        >>
                        \oneVoice
                        \afterGrace
                        f'4
                        {
                            fs'16
                        }
                    }
                }

            >>> for component in abjad.select.components(staff):
            ...     parentage = abjad.get.parentage(component)
            ...     print(f"{repr(component):30} {repr(parentage.parent)}")
            Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }") None
            Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice') Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("c'4")                    Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
            BeforeGraceContainer("cs'16")  Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
            Note("cs'16")                  BeforeGraceContainer("cs'16")
            Note("d'4")                    Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
            Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }") Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
            OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16") Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
            Chord("<e' g'>16")             OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
            Note("gs'16")                  OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
            Note("a'16")                   OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
            Note("as'16")                  OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
            Voice("e'4", name='MusicVoice') Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
            Note("e'4")                    Voice("e'4", name='MusicVoice')
            Note("f'4")                    Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
            AfterGraceContainer("fs'16")   Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
            Note("fs'16")                  AfterGraceContainer("fs'16")

        """
        return self.get(n=1)

    @property
    def prolation(self) -> fractions.Fraction:
        r"""
        Gets prolation.

        ..  container:: example

            REGRESSION. Works with grace notes (and containers):

            >>> music_voice = abjad.Voice(
            ...     r"\times 2/3 { c'4 d' e' } \times 2/3 { f' g' a' }",
            ...     name="MusicVoice"
            ... )
            >>> container = abjad.BeforeGraceContainer("cs'16")
            >>> abjad.attach(container, music_voice[0][1])
            >>> obgc = abjad.on_beat_grace_container("a'8 b'", music_voice[1][:1])
            >>> container = abjad.AfterGraceContainer("fs'16")
            >>> abjad.attach(container, music_voice[1][2])
            >>> staff = abjad.Staff([music_voice])
            >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \context Voice = "MusicVoice"
                    {
                        \tuplet 3/2
                        {
                            c'4
                            \grace {
                                cs'16
                            }
                            d'4
                            e'4
                        }
                        \tuplet 3/2
                        {
                            <<
                                \context Voice = "On_Beat_Grace_Container"
                                {
                                    \set fontSize = #-3
                                    \slash
                                    \voiceOne
                                    <
                                        \tweak font-size 0
                                        \tweak transparent ##t
                                        f'
                                        a'
                                    >8
                                    [
                                    (
                                    b'8
                                    )
                                    ]
                                }
                                \context Voice = "MusicVoice"
                                {
                                    \voiceTwo
                                    f'4
                                }
                            >>
                            \oneVoice
                            g'4
                            \afterGrace
                            a'4
                            {
                                fs'16
                            }
                        }
                    }
                }

            >>> for component in abjad.select.components(staff):
            ...     parentage = abjad.get.parentage(component)
            ...     print(f"{repr(component):30} {repr(parentage.prolation)}")
            Staff("{ { 2/3 c'4 d'4 e'4 } { 2/3 { { <f' a'>8 b'8 } { f'4 } } g'4 a'4 } }") Fraction(1, 1)
            Voice("{ 2/3 c'4 d'4 e'4 } { 2/3 { { <f' a'>8 b'8 } { f'4 } } g'4 a'4 }", name='MusicVoice') Fraction(1, 1)
            Tuplet('3:2', "c'4 d'4 e'4")   Fraction(2, 3)
            Note("c'4")                    Fraction(2, 3)
            BeforeGraceContainer("cs'16")  Fraction(2, 3)
            Note("cs'16")                  Fraction(2, 3)
            Note("d'4")                    Fraction(2, 3)
            Note("e'4")                    Fraction(2, 3)
            Tuplet('3:2', "{ { <f' a'>8 b'8 } { f'4 } } g'4 a'4") Fraction(2, 3)
            Container("{ <f' a'>8 b'8 } { f'4 }") Fraction(2, 3)
            OnBeatGraceContainer("<f' a'>8 b'8") Fraction(2, 3)
            Chord("<f' a'>8")              Fraction(2, 3)
            Note("b'8")                    Fraction(2, 3)
            Voice("f'4", name='MusicVoice') Fraction(2, 3)
            Note("f'4")                    Fraction(2, 3)
            Note("g'4")                    Fraction(2, 3)
            Note("a'4")                    Fraction(2, 3)
            AfterGraceContainer("fs'16")   Fraction(2, 3)
            Note("fs'16")                  Fraction(2, 3)

        """
        prolations = [fractions.Fraction(1)] + self._prolations()
        products = _math.cumulative_products(prolations)
        return products[-1]

    @property
    def root(self) -> _score.Component:
        r"""
        Gets root.

        ..  container:: example

            REGRESSION. Works with grace notes (and containers):

            >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
            >>> container = abjad.BeforeGraceContainer("cs'16")
            >>> abjad.attach(container, music_voice[1])
            >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
            >>> abjad.attach(abjad.Articulation(">"), obgc[0])
            >>> container = abjad.AfterGraceContainer("fs'16")
            >>> abjad.attach(container, music_voice[3])
            >>> staff = abjad.Staff([music_voice])
            >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \context Voice = "MusicVoice"
                    {
                        c'4
                        \grace {
                            cs'16
                        }
                        d'4
                        <<
                            \context Voice = "On_Beat_Grace_Container"
                            {
                                \set fontSize = #-3
                                \slash
                                \voiceOne
                                <
                                    \tweak font-size 0
                                    \tweak transparent ##t
                                    e'
                                    g'
                                >16
                                - \accent
                                [
                                (
                                gs'16
                                a'16
                                as'16
                                )
                                ]
                            }
                            \context Voice = "MusicVoice"
                            {
                                \voiceTwo
                                e'4
                            }
                        >>
                        \oneVoice
                        \afterGrace
                        f'4
                        {
                            fs'16
                        }
                    }
                }

            >>> for component in abjad.select.components(staff):
            ...     parentage = abjad.get.parentage(component)
            ...     print(f"{repr(component):30} {repr(parentage.root)}")
            Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }") Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice') Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("c'4")                    Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            BeforeGraceContainer("cs'16")  Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("cs'16")                  Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("d'4")                    Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }") Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16") Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Chord("<e' g'>16")             Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("gs'16")                  Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("a'16")                   Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("as'16")                  Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Voice("e'4", name='MusicVoice') Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("e'4")                    Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("f'4")                    Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            AfterGraceContainer("fs'16")   Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("fs'16")                  Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")

        """
        root = self.get(n=-1)
        assert isinstance(root, _score.Component), repr(root)
        return root

    ### PUBLIC METHODS ###

    def count(self, prototype=None) -> int:
        r"""
        Gets number of ``prototype`` in parentage.

        ..  container:: example

            Gets tuplet count:

            >>> staff = abjad.Staff(
            ...     r"\times 2/3 { c'2 \times 2/3 { d'8 e' f' } } \times 2/3 { c'4 d' e' }"
            ... )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \tuplet 3/2
                    {
                        c'2
                        \tuplet 3/2
                        {
                            d'8
                            e'8
                            f'8
                        }
                    }
                    \tuplet 3/2
                    {
                        c'4
                        d'4
                        e'4
                    }
                }

            >>> for component in abjad.select.components(staff):
            ...     parentage = abjad.get.parentage(component)
            ...     count = parentage.count(abjad.Tuplet)
            ...     print(f"{repr(component):55} {repr(count)}")
            Staff("{ 2/3 c'2 { 2/3 d'8 e'8 f'8 } } { 2/3 c'4 d'4 e'4 }") 0
            Tuplet('3:2', "c'2 { 2/3 d'8 e'8 f'8 }")                1
            Note("c'2")                                             1
            Tuplet('3:2', "d'8 e'8 f'8")                            2
            Note("d'8")                                             2
            Note("e'8")                                             2
            Note("f'8")                                             2
            Tuplet('3:2', "c'4 d'4 e'4")                            1
            Note("c'4")                                             1
            Note("d'4")                                             1
            Note("e'4")                                             1

        ..  container:: example

            Gets voice count:

            >>> outer_red_voice = abjad.Voice("e''8 d''", name="Red_Voice")
            >>> inner_red_voice = abjad.Voice("c''4 b' c''8", name="Red_Voice")
            >>> inner_blue_voice = abjad.Voice("e'4 f' e'8", name="Blue_Voice")
            >>> container = abjad.Container(
            ...     [inner_red_voice, inner_blue_voice],
            ...     simultaneous=True,
            ... )
            >>> outer_red_voice.append(container)
            >>> outer_red_voice.extend("d''8")
            >>> abjad.override(outer_red_voice).NoteHead.color = "#red"
            >>> command = abjad.VoiceNumber(1)
            >>> abjad.attach(command, outer_red_voice[0])
            >>> abjad.override(inner_blue_voice).NoteHead.color = "#blue"
            >>> command = abjad.VoiceNumber(2)
            >>> abjad.attach(command, inner_blue_voice[0])
            >>> dynamic = abjad.Dynamic("f")
            >>> abjad.attach(dynamic, outer_red_voice[0])
            >>> abjad.show(outer_red_voice) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(outer_red_voice)
                >>> print(string)
                \context Voice = "Red_Voice"
                \with
                {
                    \override NoteHead.color = #red
                }
                {
                    \voiceOne
                    e''8
                    \f
                    d''8
                    <<
                        \context Voice = "Red_Voice"
                        {
                            c''4
                            b'4
                            c''8
                        }
                        \context Voice = "Blue_Voice"
                        \with
                        {
                            \override NoteHead.color = #blue
                        }
                        {
                            \voiceTwo
                            e'4
                            f'4
                            e'8
                        }
                    >>
                    d''8
                }

            >>> for leaf in abjad.iterate.leaves(outer_red_voice):
            ...     depth = abjad.get.parentage(leaf).count(abjad.Voice)
            ...     print(leaf, depth)
            ...
            Note("e''8") 1
            Note("d''8") 1
            Note("c''4") 2
            Note("b'4") 2
            Note("c''8") 2
            Note("e'4") 2
            Note("f'4") 2
            Note("e'8") 2
            Note("d''8") 1

        ..  container:: example

            REGRESSION. Works with grace notes (and containers):

            >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
            >>> container = abjad.BeforeGraceContainer("cs'16")
            >>> abjad.attach(container, music_voice[1])
            >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
            >>> abjad.attach(abjad.Articulation(">"), obgc[0])
            >>> container = abjad.AfterGraceContainer("fs'16")
            >>> abjad.attach(container, music_voice[3])
            >>> staff = abjad.Staff([music_voice])
            >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \context Voice = "MusicVoice"
                    {
                        c'4
                        \grace {
                            cs'16
                        }
                        d'4
                        <<
                            \context Voice = "On_Beat_Grace_Container"
                            {
                                \set fontSize = #-3
                                \slash
                                \voiceOne
                                <
                                    \tweak font-size 0
                                    \tweak transparent ##t
                                    e'
                                    g'
                                >16
                                - \accent
                                [
                                (
                                gs'16
                                a'16
                                as'16
                                )
                                ]
                            }
                            \context Voice = "MusicVoice"
                            {
                                \voiceTwo
                                e'4
                            }
                        >>
                        \oneVoice
                        \afterGrace
                        f'4
                        {
                            fs'16
                        }
                    }
                }

            >>> for component in abjad.select.components(staff):
            ...     parentage = abjad.get.parentage(component)
            ...     count = parentage.count(abjad.Staff)
            ...     print(f"{repr(component):30} {repr(count)}")
            Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }") 1
            Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice') 1
            Note("c'4")                    1
            BeforeGraceContainer("cs'16")  1
            Note("cs'16")                  1
            Note("d'4")                    1
            Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }") 1
            OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16") 1
            Chord("<e' g'>16")             1
            Note("gs'16")                  1
            Note("a'16")                   1
            Note("as'16")                  1
            Voice("e'4", name='MusicVoice') 1
            Note("e'4")                    1
            Note("f'4")                    1
            AfterGraceContainer("fs'16")   1
            Note("fs'16")                  1

        """
        n = 0
        if prototype is None:
            prototype = _score.Component
        for component in self:
            if isinstance(component, prototype):
                n += 1
        return n

    def get(self, prototype=None, n=0) -> _score.Component | None:
        r"""
        Gets instance ``n`` of ``prototype`` in parentage.

        ..  container:: example

            >>> outer_red_voice = abjad.Voice("e''8 d''", name="Red_Voice")
            >>> inner_red_voice = abjad.Voice("c''4 b' c''8", name="Red_Voice")
            >>> inner_blue_voice = abjad.Voice("e'4 f' e'8", name="Blue_Voice")
            >>> container = abjad.Container(
            ...     [inner_red_voice, inner_blue_voice],
            ...     simultaneous=True,
            ... )
            >>> outer_red_voice.append(container)
            >>> outer_red_voice.extend("d''8")
            >>> abjad.override(outer_red_voice).NoteHead.color = "#red"
            >>> command = abjad.VoiceNumber(1)
            >>> abjad.attach(command, outer_red_voice[0])
            >>> abjad.override(inner_blue_voice).NoteHead.color = "#blue"
            >>> command = abjad.VoiceNumber(2)
            >>> abjad.attach(command, inner_blue_voice[0])
            >>> dynamic = abjad.Dynamic("f")
            >>> abjad.attach(dynamic, outer_red_voice[0])
            >>> abjad.show(outer_red_voice) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(outer_red_voice)
                >>> print(string)
                \context Voice = "Red_Voice"
                \with
                {
                    \override NoteHead.color = #red
                }
                {
                    \voiceOne
                    e''8
                    \f
                    d''8
                    <<
                        \context Voice = "Red_Voice"
                        {
                            c''4
                            b'4
                            c''8
                        }
                        \context Voice = "Blue_Voice"
                        \with
                        {
                            \override NoteHead.color = #blue
                        }
                        {
                            \voiceTwo
                            e'4
                            f'4
                            e'8
                        }
                    >>
                    d''8
                }

            ..  container:: example

                >>> leaf = abjad.get.leaf(inner_red_voice, 0)
                >>> leaf
                Note("c''4")

                >>> parentage = abjad.get.parentage(leaf)

                Returns self when ``n=0``:

                >>> parentage.get(abjad.Component, 0)
                Note("c''4")

                Returns parents with positive ``n``:

                >>> parentage.get(abjad.Component, 1)
                Voice("c''4 b'4 c''8", name='Red_Voice')

                >>> parentage.get(abjad.Component, 2)
                Container("{ c''4 b'4 c''8 } { e'4 f'4 e'8 }")

                >>> parentage.get(abjad.Component, 3)
                Voice("e''8 d''8 { { c''4 b'4 c''8 } { e'4 f'4 e'8 } } d''8", name='Red_Voice')

                Returns none with ``n`` greater than score depth:

                >>> parentage.get(abjad.Component, 4) is None
                True

                >>> parentage.get(abjad.Component, 5) is None
                True

                >>> parentage.get(abjad.Component, 99) is None
                True

                Returns score root with ``n=-1``:

                >>> parentage.get(abjad.Component, -1)
                Voice("e''8 d''8 { { c''4 b'4 c''8 } { e'4 f'4 e'8 } } d''8", name='Red_Voice')

                With other negative ``n``:

                >>> parentage.get(abjad.Component, -2)
                Container("{ c''4 b'4 c''8 } { e'4 f'4 e'8 }")

                >>> parentage.get(abjad.Component, -3)
                Voice("c''4 b'4 c''8", name='Red_Voice')

                >>> parentage.get(abjad.Component, -4)
                Note("c''4")

                Returns none for sufficiently negative ``n``:

                >>> parentage.get(abjad.Component, -5) is None
                True

                >>> parentage.get(abjad.Component, -7) is None
                True

                >>> parentage.get(abjad.Component, -99) is None
                True

            ..  container:: example

                Works with nested voices and tuplets:

                >>> leaf = abjad.get.leaf(inner_red_voice, 0)
                >>> parentage = abjad.get.parentage(leaf)
                >>> leaf
                Note("c''4")

                Nonnegative ``n``:

                >>> parentage.get(abjad.Voice, 0)
                Voice("c''4 b'4 c''8", name='Red_Voice')

                >>> parentage.get(abjad.Voice, 1)
                Voice("e''8 d''8 { { c''4 b'4 c''8 } { e'4 f'4 e'8 } } d''8", name='Red_Voice')

                >>> parentage.get(abjad.Voice, 2) is None
                True

                >>> parentage.get(abjad.Voice, 9) is None
                True

                Negative ``n``:

                >>> parentage.get(abjad.Voice, -1)
                Voice("e''8 d''8 { { c''4 b'4 c''8 } { e'4 f'4 e'8 } } d''8", name='Red_Voice')

                >>> parentage.get(abjad.Voice, -2)
                Voice("c''4 b'4 c''8", name='Red_Voice')

                >>> parentage.get(abjad.Voice, -3) is None
                True

                >>> parentage.get(abjad.Voice, -99) is None
                True

        ..  container:: example

            REGRESSION. Works with grace notes (and containers):

            >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
            >>> container = abjad.BeforeGraceContainer("cs'16")
            >>> abjad.attach(container, music_voice[1])
            >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
            >>> abjad.attach(abjad.Articulation(">"), obgc[0])
            >>> container = abjad.AfterGraceContainer("fs'16")
            >>> abjad.attach(container, music_voice[3])
            >>> staff = abjad.Staff([music_voice])
            >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \context Voice = "MusicVoice"
                    {
                        c'4
                        \grace {
                            cs'16
                        }
                        d'4
                        <<
                            \context Voice = "On_Beat_Grace_Container"
                            {
                                \set fontSize = #-3
                                \slash
                                \voiceOne
                                <
                                    \tweak font-size 0
                                    \tweak transparent ##t
                                    e'
                                    g'
                                >16
                                - \accent
                                [
                                (
                                gs'16
                                a'16
                                as'16
                                )
                                ]
                            }
                            \context Voice = "MusicVoice"
                            {
                                \voiceTwo
                                e'4
                            }
                        >>
                        \oneVoice
                        \afterGrace
                        f'4
                        {
                            fs'16
                        }
                    }
                }

            >>> for component in abjad.select.components(staff):
            ...     parentage = abjad.get.parentage(component)
            ...     result = parentage.get(abjad.Staff)
            ...     print(f"{repr(component):30} {repr(result)}")
            Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }") Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice') Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("c'4")                    Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            BeforeGraceContainer("cs'16")  Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("cs'16")                  Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("d'4")                    Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }") Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16") Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Chord("<e' g'>16")             Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("gs'16")                  Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("a'16")                   Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("as'16")                  Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Voice("e'4", name='MusicVoice') Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("e'4")                    Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("f'4")                    Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            AfterGraceContainer("fs'16")   Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
            Note("fs'16")                  Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")

        """
        if prototype is None:
            prototype = (_score.Component,)
        if not isinstance(prototype, tuple):
            prototype = (prototype,)
        if 0 <= n:
            i = 0
            for component in self:
                if isinstance(component, prototype):
                    if i == n:
                        return component
                    i += 1
        else:
            i = -1
            for component in reversed(self):
                if isinstance(component, prototype):
                    if i == n:
                        return component
                    i -= 1
        return None

    def logical_voice(self) -> dict:
        r"""
        Gets logical voice.

        ..  container:: example

            Gets logical voice of note:

            >>> voice = abjad.Voice("c'4 d'4 e'4 f'4", name="MusicVoice")
            >>> staff = abjad.Staff([voice], name="Music_Staff")
            >>> score = abjad.Score([staff], name="Score")
            >>> abjad.show(score) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(score)
                >>> print(string)
                \context Score = "Score"
                <<
                    \context Staff = "Music_Staff"
                    {
                        \context Voice = "MusicVoice"
                        {
                            c'4
                            d'4
                            e'4
                            f'4
                        }
                    }
                >>

            >>> note = voice[0]
            >>> parentage = abjad.get.parentage(note)
            >>> parentage.logical_voice()
            {'score': "Score-'Score'", 'staff group': '', 'staff': "Staff-'Music_Staff'", 'voice': "Voice-'MusicVoice'"}

        ..  container:: example

            REGRESSION. Works with grace notes (and containers):

            >>> voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
            >>> container_1 = abjad.BeforeGraceContainer("cs'16")
            >>> abjad.attach(container_1, voice[1])
            >>> container_2 = abjad.AfterGraceContainer("fs'16")
            >>> abjad.attach(container_2, voice[3])
            >>> abjad.show(voice) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(voice)
                >>> print(string)
                \context Voice = "MusicVoice"
                {
                    c'4
                    \grace {
                        cs'16
                    }
                    d'4
                    e'4
                    \afterGrace
                    f'4
                    {
                        fs'16
                    }
                }

            >>> abjad.get.parentage(voice).logical_voice()
            {'score': '', 'staff group': '', 'staff': '', 'voice': "Voice-'MusicVoice'"}

            >>> abjad.get.parentage(container_1).logical_voice()
            {'score': '', 'staff group': '', 'staff': '', 'voice': "Voice-'MusicVoice'"}

            >>> abjad.get.parentage(container_1[0]).logical_voice()
            {'score': '', 'staff group': '', 'staff': '', 'voice': "Voice-'MusicVoice'"}

            >>> abjad.get.parentage(container_2).logical_voice()
            {'score': '', 'staff group': '', 'staff': '', 'voice': "Voice-'MusicVoice'"}

            >>> abjad.get.parentage(container_2[0]).logical_voice()
            {'score': '', 'staff group': '', 'staff': '', 'voice': "Voice-'MusicVoice'"}

        """
        keys = ("score", "staff group", "staff", "voice")
        logical_voice = dict.fromkeys(keys, "")
        for component in self:
            if isinstance(component, _score.Voice):
                if not logical_voice["voice"]:
                    logical_voice["voice"] = self._id_string(component)
            elif isinstance(component, _score.Staff):
                if not logical_voice["staff"]:
                    logical_voice["staff"] = self._id_string(component)
                    # explicit staff demands a nested voice:
                    # if no explicit voice has been found,
                    # create implicit voice here with random integer
                    if not logical_voice["voice"]:
                        logical_voice["voice"] = str(id(component))
            elif isinstance(component, _score.StaffGroup):
                if not logical_voice["staff group"]:
                    logical_voice["staff group"] = self._id_string(component)
            elif isinstance(component, _score.Score):
                if not logical_voice["score"]:
                    logical_voice["score"] = self._id_string(component)
        logical_voice_ = dict(logical_voice)
        return logical_voice_

    def score_index(self) -> tuple[int | str, ...]:
        r"""
        Gets score index.

        ..  container:: example

            >>> staff_1 = abjad.Staff(r"\times 2/3 { c''2 b'2 a'2 }")
            >>> staff_2 = abjad.Staff("c'2 d'2")
            >>> score = abjad.Score([staff_1, staff_2])
            >>> abjad.show(score) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(score)
                >>> print(string)
                \new Score
                <<
                    \new Staff
                    {
                        \tuplet 3/2
                        {
                            c''2
                            b'2
                            a'2
                        }
                    }
                    \new Staff
                    {
                        c'2
                        d'2
                    }
                >>

            >>> for component in abjad.select.components(score):
            ...     parentage = abjad.get.parentage(component)
            ...     component, parentage.score_index()
            ...
            (Score("{ { 2/3 c''2 b'2 a'2 } } { c'2 d'2 }", simultaneous=True), ())
            (Staff("{ 2/3 c''2 b'2 a'2 }"), (0,))
            (Tuplet('3:2', "c''2 b'2 a'2"), (0, 0))
            (Note("c''2"), (0, 0, 0))
            (Note("b'2"), (0, 0, 1))
            (Note("a'2"), (0, 0, 2))
            (Staff("c'2 d'2"), (1,))
            (Note("c'2"), (1, 0))
            (Note("d'2"), (1, 1))

            Score root sets score index to ``()``.

        ..  container:: example

            REGRESSION. Works with grace notes (and containers):

            >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
            >>> container = abjad.BeforeGraceContainer("cs'16")
            >>> abjad.attach(container, music_voice[1])
            >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
            >>> abjad.attach(abjad.Articulation(">"), obgc[0])
            >>> container = abjad.AfterGraceContainer("fs'16")
            >>> abjad.attach(container, music_voice[3])
            >>> staff = abjad.Staff([music_voice])
            >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \context Voice = "MusicVoice"
                    {
                        c'4
                        \grace {
                            cs'16
                        }
                        d'4
                        <<
                            \context Voice = "On_Beat_Grace_Container"
                            {
                                \set fontSize = #-3
                                \slash
                                \voiceOne
                                <
                                    \tweak font-size 0
                                    \tweak transparent ##t
                                    e'
                                    g'
                                >16
                                - \accent
                                [
                                (
                                gs'16
                                a'16
                                as'16
                                )
                                ]
                            }
                            \context Voice = "MusicVoice"
                            {
                                \voiceTwo
                                e'4
                            }
                        >>
                        \oneVoice
                        \afterGrace
                        f'4
                        {
                            fs'16
                        }
                    }
                }

            >>> for component in abjad.select.components(staff):
            ...     parentage = abjad.get.parentage(component)
            ...     score_index = parentage.score_index()
            ...     print(f"{repr(component):30} {repr(score_index)}")
            Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }") ()
            Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice') (0,)
            Note("c'4")                    (0, 0)
            BeforeGraceContainer("cs'16")  (0, 0, 1, '-G')
            Note("cs'16")                  (0, 0, 1, '-G', 0)
            Note("d'4")                    (0, 1)
            Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }") (0, 2)
            OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16") (0, 2, 0)
            Chord("<e' g'>16")             (0, 2, 0, 0)
            Note("gs'16")                  (0, 2, 0, 1)
            Note("a'16")                   (0, 2, 0, 2)
            Note("as'16")                  (0, 2, 0, 3)
            Voice("e'4", name='MusicVoice') (0, 2, 1)
            Note("e'4")                    (0, 2, 1, 0)
            Note("f'4")                    (0, 3)
            AfterGraceContainer("fs'16")   (0, 0, 3, '+G')
            Note("fs'16")                  (0, 0, 3, '+G', 0)

        """
        result: list[int | str] = []
        current = self[0]
        for parent in self[1:]:
            if isinstance(current, _score.BeforeGraceContainer):
                tuple_ = type(self)(current._main_leaf).score_index()
                list_ = list(tuple_) + ["-G"]
                result[0:0] = list_
            elif isinstance(current, _score.AfterGraceContainer):
                tuple_ = type(self)(current._main_leaf).score_index()
                list_ = list(tuple_) + ["+G"]
                result[0:0] = list_
            else:
                index = parent.index(current)
                result.insert(0, index)
            current = parent
        return tuple(result)
